// Splay-  


#include<stdio.h>
//#include<memory>
#include<algorithm>
//#include<assert.h>


class SplayTree
{
	//!    .
	//!  ,    
	//! LR -     ,  .
	enum State{
		undef=0,			//!<  
		L=1,
		R=2,
		LL=(L<<2)+L, 
		LR=(L<<2)+R,
		RL=(R<<2)+L,
		RR=(R<<2)+R,
		idle=L|R,			//!<  
	};
public:
	typedef unsigned int value_type;	//!    
	//!  
	struct Node{
		Node *p1, *p2;
		value_type val;
		Node(value_type val) :val(val), p1(0), p2(0) {}
	};

	SplayTree() :treeRoot(0) {}
	~SplayTree() { Clear(treeRoot); }
	bool Insert (value_type);			//!< 
	bool Remove (value_type);			//!< 
	bool FindPassive(value_type) const;	//!< 
	bool FindAdjust (value_type);		//!<   
protected:
	Node *treeRoot;
	value_type GetMax(Node*&);			//!<    
	bool PickUp(value_type val);		//!<     
private:
	void  Clear(Node*);
	Node *RotateL(Node*) const;	// 
	Node *RotateR(Node*) const;
	State StateL (State) const;	//    
	State StateR (State) const;

	State Insert(Node*&, value_type);
	State Remove(Node*&, value_type);

	State PickUp(Node*&, value_type);		//!    
	State Splay (Node*&, State);			//!  
	void  SplayRoot(Node*&,State);			//!   
};


class SplayTreeEx: public SplayTree
{
public:
	void Join(SplayTreeEx &tree1, SplayTreeEx &tree2)
	{
		if (!tree1.treeRoot)
		{
			std::swap(treeRoot, tree2.treeRoot);
			return;
		}
		value_type val = GetMax(tree1.treeRoot);
		treeRoot = new Node(val);
		std::swap(treeRoot->p1,tree1.treeRoot);
		std::swap(treeRoot->p2, tree2.treeRoot);
	}
	bool Split(SplayTreeEx &tree1, SplayTreeEx &tree2,value_type val)
	{
		if(PickUp(val)) return true;
		std::swap(treeRoot->p1, tree1.treeRoot);
		std::swap(treeRoot->p2, tree2.treeRoot);
		return false;
	}
};


SplayTree::Node* SplayTree::RotateL(Node *node) const
{
	Node* p2 = node->p2;
	node->p2 = p2->p1;
	p2->p1 = node;
	return p2;
}

SplayTree::Node* SplayTree::RotateR(Node *node) const
{
	Node* p1 = node->p1;
	node->p1 = p1->p2;
	p1->p2 = node;
	return p1;
}

SplayTree::State SplayTree::StateL(State st) const
{
	return st == idle ? st : State(L | st << 2);
}

SplayTree::State SplayTree::StateR(State st) const
{
	return st == idle ? st : State(R | st << 2);
}

void SplayTree::Clear(Node *node)
{
	if (!node) return;
	Clear(node->p1);
	Clear(node->p2);
	delete node;
}


void SplayTree::SplayRoot(Node *&node, State st)
{
	switch (st)
	{
	case L: node = RotateR(node); break;
	case R: node = RotateL(node); break;
	}
}


SplayTree::State SplayTree::Splay(Node *&node, State st)
{
	switch (st)
	{
	case LL: node = RotateR(node); break;
	case RR: node = RotateL(node);  break;
	case RL: node->p1 = RotateL(node->p1); break;
	case LR: node->p2 = RotateR(node->p2); break;
	default: return st;
	}
	st = State(st&(R|L));
	SplayRoot(node, st);
	return undef;
}

// true    
bool SplayTree::PickUp(value_type val)
{
	State st = PickUp(treeRoot, val);
	if (st == idle) return true;
	SplayRoot(treeRoot, st);
	return false;
}

SplayTree::State SplayTree::PickUp(Node *&node, value_type val)
{
	
	if (!node) return idle;
	if (val == node->val) return undef;
	State st=(val < node->val)?
		StateL(PickUp(node->p1,val)):
		StateR(PickUp(node->p2,val));
	return Splay(node,st);
}

SplayTree::value_type SplayTree::GetMax(Node *&node)
{
	if (!node->p2)
	{
		std::auto_ptr<Node> obj(node);
		node = node->p1;
		return obj->val;
	}
	return GetMax(node->p2);
}


// true   
bool SplayTree::Insert(value_type val)
{
	State st = Insert(treeRoot, val);
	SplayRoot(treeRoot, st);
	return st == idle;
}

SplayTree::State SplayTree::Insert(Node *&node, value_type val)
{
	if (!node) {
		node = new Node(val);
		return undef;
	}
	if (val == node->val) return idle;
	State st = (val < node->val)?
		StateL(Insert(node->p1, val)):
		StateR(Insert(node->p2, val));
	return Splay(node, st);
}


// true    
bool SplayTree::Remove(value_type val)
{
	if (PickUp(val)) return true;
	if (treeRoot->p1) treeRoot->val = GetMax(treeRoot->p1);
	else{
		std::auto_ptr<Node> ref(treeRoot);
		treeRoot = treeRoot->p2;
	}
	return false;
}


// true   
bool SplayTree::FindPassive(value_type val) const
{
	for (Node *node = treeRoot; node;)
	{
		if (val == node->val) return false;
		node = (val < node->val) ? node->p1 : node->p2;
	}
	return true;
}

// true  
bool SplayTree::FindAdjust(value_type val)
{
	return !PickUp(val);
}




int main()
{
	SplayTree tree;
	for (int p = 0; p<1000; p++)
	{
		tree.Insert(rand() % 100);
		tree.Remove(rand() % 100);
	}
	return 0;
}


